class PyTerrierIntegration():
    def __init__(self, tira_client):
        self.tira_client = tira_client
        self.irds_docker_image = 'webis/tira-application:0.0.36'

    def retriever(self, approach, dataset=None, verbose=False):
        from tira.pyterrier_util import TiraFullRankTransformer
        input_dir = self.ensure_dataset_is_cached(dataset, dataset)
        return TiraFullRankTransformer(approach, self.tira_client, input_dir, verbose)

    def ensure_dataset_is_cached(self, irds_dataset_id, dataset):
        import os
        from pathlib import Path
        from tira.io_utils import run_cmd
        import json

        cache_dir = self.tira_client.tira_cache_dir + '/pyterrier/' + irds_dataset_id
        full_rank_data = cache_dir + '/full-rank/'
        truth_data = cache_dir + '/truth-data/'
        irds_cache = cache_dir + '/irds-cache/'

        if os.path.isfile(full_rank_data + '/documents.jsonl'):
            return full_rank_data

        Path(full_rank_data).mkdir(parents=True, exist_ok=True)
        Path(truth_data).mkdir(parents=True, exist_ok=True)
        Path(irds_cache).mkdir(parents=True, exist_ok=True)

        run_cmd(['docker', 'run',
                 '-v', irds_cache + ':/root/.ir_datasets/:rw',
                 '-v', full_rank_data + ':/output/:rw',
                 '-v', truth_data + ':/truth/:rw',
                 '--entrypoint', '/irds_cli.sh',
                 self.irds_docker_image,
                 '--output_dataset_path', '/output',
                 '--ir_datasets_id', irds_dataset_id,
                 '--output_dataset_truth_path', '/truth'
                 ])

        return full_rank_data      

    def create_rerank_file(self, run_df=None, run_file=None, irds_dataset_id=None):
        from pathlib import Path
        import tempfile
        from tira.io_utils import run_cmd
        import gzip
        import json
        from tira.third_party_integrations import persist_and_normalize_run

        if run_df is None and run_file is None:
            raise ValueError('Please pass either run_df or run_file')

        if run_file is not None:
            return run_file

        run_file = tempfile.TemporaryDirectory('-rerank-run').name
        Path(run_file).mkdir(parents=True, exist_ok=True)

        if 'text' not in run_df.columns and 'body' not in run_df.columns:
            if not irds_dataset_id:
                raise ValueError(f'Please pass a irds_dataset_id. Got {irds_dataset_id}.')
            persist_and_normalize_run(run_df, 'system-is-ignored', run_file)

            cache_dir = self.tira_client.tira_cache_dir + '/pyterrier/' + irds_dataset_id
            irds_cache = cache_dir + '/irds-cache/'

            run_cmd(['docker', 'run',
                     '-v', irds_cache + ':/root/.ir_datasets/:rw',
                     '-v', run_file + ':/output/:rw',
                     '--entrypoint', '/irds_cli.sh',
                     self.irds_docker_image,
                     '--output_dataset_path', '/output',
                     '--ir_datasets_id', irds_dataset_id,
                     '--rerank', '/output'
            ])
        else:
            with gzip.open(run_file + '/rerank.jsonl.gz', 'wt') as f:
                for _, i in run_df.iterrows():
                    i = i.to_dict()

                    for k in ['original_query', 'original_document']:
                        if k not in i:
                            i[k] = {}

                    if 'text' not in i and 'body' in i:
                        i['text'] = i['body']

                    if 'text' not in i:
                        raise ValueError(f'I expect a field "text", but only found fields {i.keys()}.')

                    f.write(json.dumps(i) + '\n')

        return run_file

    def from_submission(self, approach, dataset=None, datasets=None):
        software = self.tira_client.docker_software(approach)
        
        if not 'ir_re_ranker' in software or not software['ir_re_ranker']:
            return self.from_retriever_submission(approach, dataset, datasets=datasets)
        else:
            from tira.pyterrier_util import TiraRerankingTransformer
            
            return TiraRerankingTransformer(approach, self.tira_client, dataset, datasets)

    def from_retriever_submission(self, approach, dataset, previous_stage=None, datasets=None):
        import pyterrier as pt
        import pandas as pd
        from tira.ir_datasets_util import translate_irds_id_to_tirex
        task, team, software = approach.split('/')

        if dataset and datasets:
            raise ValueError(f'You can not pass both, dataset and datasets. Got dataset = {dataset} and datasets= {datasets}')

        if not datasets:
            datasets = [dataset]

        df_ret = []
        for dataset in datasets:
            ret, run_id = self.tira_client.download_run(task, translate_irds_id_to_tirex(dataset), software, team, previous_stage, return_metadata=True)
            ret['qid'] = ret['query'].astype(str)
            ret['docno'] = ret['docid'].astype(str)
            del ret['query']
            del ret['docid']

            ret['tira_task'] = task
            ret['tira_dataset'] = dataset
            ret['tira_first_stage_run_id'] = run_id
            df_ret += [ret]

        return pt.Transformer.from_df(pd.concat(df_ret))

    def transform_queries(self, approach, dataset, file_selection= '/*.jsonl'):
        from pyterrier.apply import generic
        import pandas as pd
        from glob import glob
        from tira.ir_datasets_util import translate_irds_id_to_tirex
        glob_entry = self.tira_client.get_run_output(approach, translate_irds_id_to_tirex(dataset)) + file_selection
        matching_files = glob(glob_entry)
        if len(matching_files) == 0:
            raise ValueError(f'Could not find a matching query output. Found: {matching_files}. Please specify the file_selection to resolve this.')

        ret = pd.read_json(matching_files[0], lines=True, dtype={'qid': str, 'query': str, 'query_id': str})
        if 'qid' not in ret and 'query_id' in ret:
            ret['qid'] = ret['query_id']
            del ret['query_id']

        cols = [i for i in ret.columns if i not in ['qid']]
        ret = {str(i['qid']): i for _, i in ret.iterrows()}

        def __transform_df(df):
            for col in cols:
                df[col] = df['qid'].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    def transform_documents(self, approach, dataset, file_selection= '/*.jsonl'):
        from pyterrier.apply import generic
        import pandas as pd
        from glob import glob
        from tira.ir_datasets_util import translate_irds_id_to_tirex
        glob_entry = self.tira_client.get_run_output(approach, translate_irds_id_to_tirex(dataset)) + file_selection
        matching_files = glob(glob_entry)
        if len(matching_files) == 0:
            raise ValueError('Could not find a matching document output. Found: ' + matching_files + '. Please specify the file_selection to resolve this.')

        ret = pd.read_json(matching_files[0], lines=True)
        cols = [i for i in ret.columns if i not in ['docno']]
        ret = {str(i['docno']): i for _, i in ret.iterrows()}

        def __transform_df(df):
            for col in cols:
                df[col] = df['docno'].apply(lambda i: ret[str(i)][col])
            return df

        return generic(__transform_df)

    def reranker(self, approach, irds_id=None):
        from tira.pyterrier_util import TiraLocalExecutionRerankingTransformer
        return TiraLocalExecutionRerankingTransformer(approach, self.tira_client, irds_id=irds_id)




from tira.third_party_integrations import persist_and_normalize_run
import pandas as pd

def test_variant_01():
    run = pd.DataFrame([{"qid": "1", "score": "1", "docno": "d1"},  {"qid": "1", "score": "0", "docno": "d2"},  {"qid": "1", "score": "2", "docno": "d3"}])
    persist_and_normalize_run(run, 'system_name', default_output='/tmp/variant_01.run.txt')
    
def test_variant_02():
    run = pd.DataFrame([{"qid": "1", "score": "1", "docno": "d1"},  {"qid": "1", "score": "0", "docno": "d2"},  {"qid": "1", "score": "2", "docno": "d3"}])
    persist_and_normalize_run(run, 'system_name', output_file='/tmp/variant_02.run.txt')

def test_variant_03():
    run = pd.DataFrame([{"qid": "1", "score": "1", "docno": "d1"},  {"qid": "1", "score": "0", "docno": "d2"},  {"qid": "1", "score": "2", "docno": "d3"}])
    persist_and_normalize_run(run, 'system_name')




from tira.local_client import Client

import pandas as pd


def test_for_loading_keyphrase_extraction_for_multiple_documents():
    expected = {
        'FT921-3964': ["increased economic aid","king hussein","increase aid","president george bush","tariq aziz iraq's deputy prime minister"],
        'LA061489-0137': ["public appearances gorbachev","crowds","government ministries gorbachev","place gorbachev","west german well-wishers"]}
    
    actual = keyphrase_extraction(['FT921-3964', 'LA061489-0137'])
    
    assert len(actual) == 2
    assert actual['FT921-3964'] == expected['FT921-3964']
    assert actual['LA061489-0137'] == expected['LA061489-0137']


def test_for_loading_keyphrase_extraction_for_single_documents():
    expected = {"LA120390-0047": ["even friendly technology takeovers","largest technology takeover","computer industry takeover","statement sunday","attempts"]}
    
    actual = keyphrase_extraction(["LA120390-0047"])
    
    assert len(actual) == 1
    assert actual["LA120390-0047"] == expected["LA120390-0047"]


def test_pyterrier_can_be_loaded():
    from tira.third_party_integrations import ensure_pyterrier_is_loaded
    ensure_pyterrier_is_loaded()


def keyphrase_extraction(docs):
    queries = pd.DataFrame([{'docno': str(i)} for i in docs])
    tira = Client('tests/resources/')
    query_segmentation =  tira.pt.transform_documents('ir-benchmarks/webis-keyphrase-extraction/BCExtractorFO', dataset='d1')
    
    ret = query_segmentation(queries)
    return {i['docno']: i['keyphrases'] for _, i in ret.iterrows()}




from tira.rest_api_client import Client
tira = Client()


def test_all_softwares_works_for_tirex():
    actual = tira.all_softwares("ir-benchmarks")
    
    assert 'ir-benchmarks/tira-ir-starter/ChatNoir' in actual
    assert 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)' in actual

def test_details_of_public_software_01():
    actual = tira.docker_software( 'ir-benchmarks/tira-ir-starter/TF_IDF (tira-ir-starter-pyterrier)')

    assert actual is not None
    assert actual['ir_re_ranker'] == False

def test_output_of_run():
    actual = tira.get_run_output('ir-benchmarks/tira-ir-starter/BM25 Re-Rank (tira-ir-starter-pyterrier)', 'antique-test-20230107-training')
    actual = open(actual + '/run.txt', 'r').read()
    
    assert actual.startswith('8293 Q0 8293_1 1 33.568841093129976 pyterrier.default_pipelines.wmodel_text_scorer')




from tira.local_execution_integration import LocalExecutionIntegration


def test_non_existing_notebook_1():
    notebook = 'does-not-exist'

    actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

    assert actual is None


def test_non_existing_notebook_2():
    notebook = 'does-not-exist/1/does-not-exist.ipynb'

    actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

    assert actual is None


def test_notebook_submission_without_previous_stages():
    notebook = 'tests/resources/pyterrier-notebook-without-previous-stages.ipynb'

    expected = 'TIRA_COMMAND=/workspace/run-pyterrier-notebook.py --input ${TIRA_INPUT_DIRECTORY} --output ${TIRA_OUTPUT_DIRECTORY} --notebook /workspace/pyterrier-notebook-without-previous-stages.ipynb'
    actual = LocalExecutionIntegration().export_submission_from_jupyter_notebook(notebook)

    assert expected == actual




from tira.local_execution_integration import LocalExecutionIntegration as Client
import tempfile

def test_export_of_file_from_bash_image():
    tira = Client()
    local_file = tempfile.NamedTemporaryFile().name
    tira.export_file_from_software('/etc/issue', local_file, image='bash:alpine3.16')
    
    actual = open(local_file).read()
    expected = '''Welcome to Alpine Linux 3.16
Kernel \\r on an \\m (\\l)

'''
    assert actual == expected

def test_export_of_file():
    tira = Client()
    local_file = tempfile.NamedTemporaryFile().name
    tira.export_file_from_software('/etc/alpine-release', local_file, image='bash:alpine3.16')
    
    actual = open(local_file).read()
    expected = '''3.16.5\n'''
    assert actual == expected

def test_extraction_of_entrypoint():
    expected = 'docker-entrypoint.sh'

    tira = Client()
    actual = tira.extract_entrypoint(image='bash:alpine3.16')

    assert actual == expected




#!/usr/bin/env python3
import argparse
import os
import sys
import subprocess

from tira.inference_server import run_inference_server


def parse_args():
    parser = argparse.ArgumentParser(prog='tira-run-inference-server', description='')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--script', type=str, help='The python script containing the model and the predict function.', required=False)
    group.add_argument('--notebook', type=str, help='The notebook containing the model and the predict function.', required=False)

    parser.add_argument('--port', type=int, help='Port number of local server.', required=True)
    parser.add_argument('--log', type=str, help='Server log level. One of [INFO, WARNING, DEBUG]. Default is INFO', default='INFO', required=False)

    return parser.parse_args()


def main():
    args = parse_args()

    if not (5000 <= args.port <= 65536):
        print(f"Invalid port number {args.port}. Must be in range 5000 - 65536.")
        sys.exit(-1)

    if args.notebook is not None:
        notebook_name = args.notebook
        # convert notebook to python script
        command_convert = f"jupyter nbconvert --to script {notebook_name}"
        try:
            return_code = subprocess.check_call(command_convert, shell=True, stdout=sys.stdout, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            return_code = 2
        if return_code != 0:
            print(f"Exception while converting notebook '{notebook_name}'. Stopping server.")
            sys.exit(return_code)
        module_name = notebook_name
    else:
        module_name = args.script

    absolute_path = os.path.splitext(os.path.abspath(module_name))[0] + '.py'
    module_name = os.path.splitext(module_name)[0].replace('/', '.')

    # flag execution for running as inference server (see third_party_integrations.is_running_as_inference_server)
    os.environ['TIRA_INFERENCE_SERVER'] = 'True'

    run_inference_server(base_module=module_name, absolute_path=absolute_path, internal_port=args.port, loglevel=args.log)



from tira.local_client import Client

import pandas as pd


def test_for_loading_query_segmentations_for_multiple_queries():
    expected = {'350': ["health and","computer terminals"], '353': ["antarctica","exploration"]}
    
    actual = query_segmentation([350, 353])
    
    assert len(actual) == 2
    assert actual['350'] == expected['350']
    assert actual['353'] == expected['353']


def test_for_loading_query_segmentations_for_single_query():
    expected = {'351': ["falkland","petroleum exploration"]}
    
    actual = query_segmentation([351])
    
    assert len(actual) == 1
    assert actual['351'] == expected['351']


def test_pyterrier_can_be_loaded():
    from tira.third_party_integrations import ensure_pyterrier_is_loaded
    ensure_pyterrier_is_loaded()

def test_rest_query_submission_with_rest_api_01():
    from tira.rest_api_client import Client
    tira = Client()

    q = tira.pt.transform_queries('ir-benchmarks/ows/query-segmentation-wt-snp', 'disks45-nocr-trec-robust-2004-20230209-training')
    assert len(q(pd.DataFrame([{'qid': '303'}]))) == 1
    assert q(pd.DataFrame([{'qid': '303'}])).iloc[0].to_dict()['segmentation'] == ['hubble telescope achievements']

def test_rest_query_submission_with_rest_api_02():
    from tira.rest_api_client import Client
    tira = Client()

    q = tira.pt.transform_queries('ir-benchmarks/ows/query-segmentation-hyb-a', 'disks45-nocr-trec-robust-2004-20230209-training')
    assert len(q(pd.DataFrame([{'qid': '303'}]))) == 1
    assert q(pd.DataFrame([{'qid': '303'}])).iloc[0].to_dict()['segmentation'] == ['hubble telescope', 'achievements']

def test_rest_query_submission_with_rest_api_03():
    from tira.rest_api_client import Client
    from tira.third_party_integrations import ensure_pyterrier_is_loaded
    import pyterrier as pt
    ensure_pyterrier_is_loaded()
    tira = Client()

    q = tira.pt.transform_queries('ir-benchmarks/fschlatt/matching-taping', pt.get_dataset("irds:clueweb09/en/trec-web-2009"))
    assert len(q(pd.DataFrame([{'qid': '26'}]))) == 1
    assert q(pd.DataFrame([{'qid': '26'}])).iloc[0].to_dict()['mean_health_score'] == 168.416
    assert q(pd.DataFrame([{'qid': '26'}])).iloc[0].to_dict()['median_health_score'] == 130.3137

def test_rest_query_submission_with_rest_api_04():
    from tira.rest_api_client import Client
    from tira.third_party_integrations import ensure_pyterrier_is_loaded
    import pyterrier as pt
    ensure_pyterrier_is_loaded()
    tira = Client()

    q = tira.pt.transform_queries('ir-benchmarks/qpptk/fixed-sealer', pt.get_dataset("irds:disks45/nocr/trec-robust-2004"))
    assert len(q(pd.DataFrame([{'qid': '301'}]))) == 1
    assert q(pd.DataFrame([{'qid': '301'}])).iloc[0].to_dict()['max-idf'] == 3.5629408815
    assert q(pd.DataFrame([{'qid': '301'}])).iloc[0].to_dict()['avg-idf'] == 2.4740487603

def query_segmentation(queries):
    queries = pd.DataFrame([{'qid': str(i)} for i in queries])
    tira = Client('tests/resources/')
    query_segmentation =  tira.pt.transform_queries('ir-benchmarks/webis-query-segmentation/wt-snp-baseline', dataset='d1')
    
    ret = query_segmentation(queries)
    return {i['qid']: i['segmentation'] for _, i in ret.iterrows()}




import requests
import json
import pandas as pd
import os
import zipfile
import io
import docker
import time
from random import random
from tira.pyterrier_integration import PyTerrierIntegration
from tira.local_execution_integration import LocalExecutionIntegration


class Client():
    def __init__(self, api_key=None, failsave_retries=5, failsave_max_delay=15, api_user_name=None):
        self.tira_cache_dir = os.environ.get('TIRA_CACHE_DIR', os.path.expanduser('~') + '/.tira')
        self.json_cache = {}

        if api_key is None:
            self.api_key = self.load_settings()['api_key']
            self.api_user_name = self.load_settings()['api_user_name']
        else:
            self.api_key = api_key
            self.api_user_name = api_user_name

        self.failsave_retries = 1
        if self.api_key != 'no-api-key':
            self.fail_if_api_key_is_invalid()
        self.pt = PyTerrierIntegration(self)
        self.local_execution = LocalExecutionIntegration(self)

        self.failsave_retries = failsave_retries
        self.failsave_max_delay = failsave_max_delay

    def load_settings(self):
        try:
            return json.load(open(self.tira_cache_dir + '/.tira-settings.json', 'r'))
        except:
            print(f'No settings given in {self.tira_cache_dir}/.tira-settings.json. I will use defaults.')
            return {'api_key': 'no-api-key', 'api_user_name': 'no-api-key-user'}

    def fail_if_api_key_is_invalid(self):
        role = self.json_response('/api/role')
        
        if not role or 'status' not in role or 'role' not in role or 0 != role['status']:
            raise ValueError('It seems like the api key is invalid. Got: ', role)

    def datasets(self, task):
        return json.loads(self.json_response(f'/api/datasets_by_task/{task}')['context']['datasets'])

    def docker_software_id(self, approach):
        return self.docker_software(approach)['docker_software_id']

    def all_softwares(self, task_id):
        """
        Return all public submissions.
        """
        return [task_id + '/' + i['vm_id'] + '/' + i['display_name'] for i in self.json_response(f'/api/task/{task_id}/public-submissions')['context']['public_submissions']]

    def docker_software(self, approach):
        task, team, software = approach.split('/')
        return self.json_response(f'/api/task/{task}/submission-details/{team}/{software}')['context']['submission']

    def docker_software_details(self, approach):
        task, team, software = approach.split('/')
        ret = self.json_response(f'/task/{task}/vm/{team}/software_details/{software}')

        return ret

    def metadata_for_task(self, task_name, team_name):
        return self.json_response(f'/api/task/{task_name}/user/{team_name}')

    def add_docker_software(self, image, command, tira_vm_id, tira_task_id, code_repository_id, build_environment):
        headers = {   
            'Api-Key': self.api_key,
            'Api-Username': self.api_user_name,
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }
        self.fail_if_api_key_is_invalid()
        url = f'https://www.tira.io/task/{tira_task_id}/vm/{tira_vm_id}/add_software/docker'
        ret = requests.post(url, headers=headers, json={"action": "post", "image": image, "command": command, "code_repository_id": code_repository_id,"build_environment": json.dumps(build_environment)})
        ret = ret.content.decode('utf8')
        ret = json.loads(ret)

        assert ret['status'] == 0
        print(f'Software with name {ret["context"]["display_name"]} was created.')
        print(f'Please visit https://www.tira.io/submit/{tira_task_id}/user/{tira_vm_id}/docker-submission to run your software.')

    def submissions(self, task, dataset):
        response = self.json_response(f'/api/submissions/{task}/{dataset}')['context']
        ret = []
        
        for vm in response['vms']:
            for run in vm['runs']:
                if 'review' in run:
                    for k,v in run['review'].items():
                        run['review_' + k] = v
                    del run['review']

                ret += [{**{'task': response['task_id'], 'dataset': response['dataset_id'], 'team': vm['vm_id']}, **run}]

        return pd.DataFrame(ret)

    def submissions_with_evaluation_or_none(self, task, dataset, team, software):
        """ This method returns all runs of the specified software in the task on the dataset by the team.
            This is especially suitable to batch evaluate all submissions of the software because the evaluation is none if no successfull evaluation was conducted (or there is a new evaluator).
            E.g., by code like:

            ```
            for approach in ['approach-1', ..., 'approach-n]:
                runs_for_approach = tira.submissions_with_evaluation_or_none(task, dataset, team, approach)
                for i in runs_for_approach:
                    if not i['evaluation']:
                        tira.evaluate_run(team, dataset, i['run_id'])
            ```
        """
        submissions = self.submissions(task, dataset)
        evaluations = self.evaluations(task, dataset, join_submissions=False)
        run_to_evaluation = {}
        for _, i in evaluations.iterrows():
            i = i.to_dict()
            run_id = i["run_id"]
            del i['task']
            del i['dataset']
            del i['team']
            del i['run_id']
            run_to_evaluation[run_id] = i

        if len(submissions) < 1:
            return []

        submissions = submissions[(submissions['task'] == task) & (submissions['dataset'] == dataset) & (submissions['team'] == team) & (submissions['software'] == software)]
        ret = []
        for run_id in submissions.run_id.unique():
            ret += [{"run_id": run_id, "task": task, "dataset": dataset, "team": team, "software": software, "evaluation": run_to_evaluation.get(run_id, None)}]

        return ret

    def evaluations(self, task, dataset, join_submissions=True):
        response = self.json_response(f'/api/evaluations/{task}/{dataset}')['context']
        ret = []
        evaluation_keys = response['ev_keys']

        if join_submissions:
            runs_to_join = {}
            for _, i in self.submissions(task, dataset).iterrows():
                i = i.to_dict()
                runs_to_join[(i['team'], i['run_id'])] = {'software': i['software'], 'input_run_id': i['input_run_id'], 'is_upload': i['is_upload'], 'is_docker': i['is_docker']}

        for evaluation in response['evaluations']:
            run = {'task': response['task_id'], 'dataset': response['dataset_id'], 'team': evaluation['vm_id'], 'run_id': evaluation['input_run_id'], 'evaluation_run_id': evaluation['run_id'], 'published': evaluation['published'], 'blinded': evaluation['blinded']}

            if join_submissions and (run['team'], run['run_id']) in runs_to_join:
                software = runs_to_join[(run['team'], run['run_id'])]
                for k,v in software.items():
                    run[k] = v

            for measure_id, measure in zip(range(len(evaluation_keys)), evaluation_keys):
                run[measure] = evaluation['measures'][measure_id]

            ret += [run]

        return pd.DataFrame(ret)

    def run_was_already_executed_on_dataset(self, approach, dataset):
        return self.get_run_execution_or_none(approach, dataset) is not None

    def get_run_output(self, approach, dataset, allow_without_evaluation=False):
        """
        Downloads the run (or uses the cached version) of the specified approach on the specified dataset.
        Returns the directory containing the outputs of the run.
        """
        task, team, software = approach.split('/')
        run_execution = self.get_run_execution_or_none(approach, dataset)
        if run_execution:
            return self.download_zip_to_cache_directory(task, dataset, team, run_execution['run_id'])

        run_execution = self.submissions_with_evaluation_or_none(task, dataset, team, software)
        run_execution = [i for i in run_execution if ('evaluation' in i and i['evaluation']) or allow_without_evaluation]

        if run_execution is None or len(run_execution) < 1:
            raise ValueError(f'Could not get run for approach "{approach}" on dataset "{dataset}".')

        return self.download_zip_to_cache_directory(run_execution[0]['task'], run_execution[0]['dataset'], run_execution[0]['team'], run_execution[0]['run_id'])

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None):
        task, team, software = approach.split('/')

        public_runs = self.json_response(f'/api/list-runs/{task}/{dataset}/{team}/' + software.replace(' ', '%20'))
        if public_runs and 'context' in public_runs and 'runs' in public_runs['context'] and public_runs['context']['runs']:
            return {'task': task, 'dataset': dataset, 'team': team, 'run_id': public_runs['context']['runs'][0]}

        df_eval = self.submissions_of_team(task=task, dataset=dataset, team=team)
        if len(df_eval) <= 0:
            return None

        ret = df_eval[(df_eval['dataset'] == dataset) & (df_eval['software'] == software)]
        if len(ret) <= 0:
            return None
        
        if team:
            ret = ret[ret['team'] == team]

        if len(ret) <= 0:
            return None

        if previous_stage_run_id:
            ret = ret[ret['input_run_id'] == previous_stage_run_id]

        if len(ret) <= 0:
            return None

        return ret[['task', 'dataset', 'team', 'run_id']].iloc[0].to_dict()
        
    def download_run(self, task, dataset, software, team=None, previous_stage=None, return_metadata=False):
        ret = self.get_run_execution_or_none(f'{task}/{team}/{software}', dataset, previous_stage)
        if not ret:
            raise ValueError(f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}", software="{software}", team={team}, previous_stage={previous_stage}')
        run_id = ret['run_id']
        
        ret = self.download_zip_to_cache_directory(**ret)
        ret = pd.read_csv(ret + '/run.txt', sep='\\s+', names=["query", "q0", "docid", "rank", "score", "system"], dtype={"query": str, "docid": str})
        if return_metadata:
            return ret, run_id
        else:
            return ret

    def download_evaluation(self, task, dataset, software, team):
        ret = self.submissions_with_evaluation_or_none(task, dataset, team, software)
        if not ret or len(ret) < 1:
            raise ValueError(f'I could not find a run for the filter criteria task="{task}", dataset="{dataset}", software="{software}", team={team}.')
        run_id = ret[0]['run_id']

        submissions = self.submissions(task, dataset)
        submissions = submissions[(submissions['input_run_id'] == run_id) & (submissions['is_evaluation'])]

        if submissions is None or len(submissions) < 1:
            raise ValueError(f'I could not find a evaluation for the filter criteria task="{task}", dataset="{dataset}", software="{software}", team={team}, run_id={run_id}.')

        return self.download_zip_to_cache_directory(task, dataset, team, submissions.iloc[0].to_dict()['run_id'])

    def download_dataset(self, task, dataset, truth_dataset=False):
        """
        Download the dataset. Set truth_dataset to true to load the truth used for evaluations.
        """
        target_dir = f'{self.tira_cache_dir}/extracted_datasets/{task}/{dataset}/'
        suffix = ("input-data" if not truth_dataset else "truth-data")
        if os.path.isdir(target_dir + suffix):
            return target_dir + suffix
        data_type = 'training' if dataset.endswith('-training') else 'test'
        self.download_and_extract_zip(f'https://www.tira.io/data-download/{data_type}/input-{("" if not truth_dataset else "truth")}/{dataset}.zip', target_dir)

        os.rename(target_dir + f'/{dataset}', target_dir + suffix)

        return target_dir + suffix

    def download_zip_to_cache_directory(self, task, dataset, team, run_id):
        target_dir = f'{self.tira_cache_dir}/extracted_runs/{task}/{dataset}/{team}'

        if os.path.isdir(target_dir + f'/{run_id}'):
            return target_dir + f'/{run_id}/output'

        self.download_and_extract_zip(f'https://www.tira.io/task/{task}/user/{team}/dataset/{dataset}/download/{run_id}.zip', target_dir)

        return target_dir + f'/{run_id}/output'

    def add_run_to_leaderboard(self, task, team, dataset, evaluation_run_id=None, run_id=None):
        """
        Publish the specified run to the leaderboard.
        
        This is especially suitable to batch add all submissions of submissions, e.g., by code like:

            ```
            for approach in ['approach-1', ..., 'approach-n]:
                runs_for_approach = tira.submissions_with_evaluation_or_none(task, dataset, team, approach)
                for i in runs_for_approach:
                    if i['evaluation']:
                        tira.add_run_to_leaderboard(TASK, team, dataset, run_id=i['run_id'])
        
        """
        if run_id and evaluation_run_id:
            raise ValueError(f'Please pass either a evaluation_run_id or a run_id, but both were passed: evaluation_run_id={evaluation_run_id}, run_id={run_id}')
        if run_id and not evaluation_run_id:
            submissions = self.submissions(task, dataset)
            submissions = submissions[(submissions['input_run_id'] == run_id) & (submissions['is_evaluation'])]
        
            for evaluation_run_id in submissions['run_id'].unique():
                self.add_run_to_leaderboard(task, team, dataset, evaluation_run_id=evaluation_run_id)
    
        if evaluation_run_id:
            print(f'Publish run: {evaluation_run_id}.')
            ret = self.json_response(f'/publish/{team}/{dataset}/{evaluation_run_id}/true')

            if ('status' not in ret) or ('0' != ret['status']) or ('published' not in ret) or (not ret['published']):
                raise ValueError(f'Adding the run to the leaderboard failed. Got {ret}')

    def get_configuration_of_evaluation(self, task_id, dataset_id):
        """ Get the configuration of the evaluator for the passed dataset inside the task specified by task_id.
        """
        ret = self.json_response(f'/api/configuration-of-evaluation/{task_id}/{dataset_id}')

        if 'status' not in ret or '0' != str(ret['status']):
            raise ValueError(f'Failed to load configuration of an evaluator. Got {ret}')

        return ret['context']['dataset']
    

    def evaluate_run(self, team, dataset, run_id):
        """ Evaluate the run of the specified team and identified by the run_id (the run must be submitted on the specified dataset).
        """
        ret = self.json_response(f'/grpc/{team}/run_eval/{dataset}/{run_id}')

        if 'status' not in ret or '0' != str(ret['status']):
            raise ValueError(f'Failed to evaluate the run. Got {ret}')

        return ret

    def download_and_extract_zip(self, url, target_dir):
        for i in range(self.failsave_retries):
            status_code = None
            try:
                headers={"Api-Key": self.api_key, "Api-Username": self.api_user_name}
                if self.api_key == 'no-api-key':
                    del headers["Api-Key"]
                if self.api_user_name == 'no-api-key-user':
                    del headers["Api-Username"]
            
                r = requests.get(url, headers=headers)
                status_code = r.status_code
                z = zipfile.ZipFile(io.BytesIO(r.content))
                z.extractall(target_dir)
                
                return
            except Exception as e:
                sleep_time = 1+int(random()*self.failsave_max_delay)
                print(e)
                print(f'Code: {status_code}')
                print(f'Error occured while fetching {url}. I will sleep {sleep_time} seconds and continue.')
                time.sleep(sleep_time)

    def get_authentication_cookie(self, user, password):
        import requests

        resp = requests.get('https://www.tira.io/session/csrf', headers={'x-requested-with': 'XMLHttpRequest'})

        header = {
            'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
            'cookie': '_forum_session=' + resp.cookies['_forum_session'],
            'x-csrf-token': resp.json()['csrf'],
            'x-requested-with': 'XMLHttpRequest'
        }

        resp = requests.post('https://www.tira.io/session', data=f'login={user}&password={password}', headers=header)

        return f'_t={resp.cookies["_t"]}; _forum_session={resp.cookies["_forum_session"]}'

    def run_software(self, approach, dataset, resources, rerank_dataset='none'):
        task, team, software = approach.split('/')
        authentication_cookie = self.get_authentication_cookie(self.load_settings()['user'], self.load_settings()['password'])
        
        software_id = self.docker_software_id(approach)
        if not software_id:
            raise ValueError(f'Could not find software id for "{approach}". Got: "{software_id}".')
        
        url = f'https://www.tira.io/grpc/{task}/{team}/run_execute/docker/{dataset}/{software_id}/{resources}/{rerank_dataset}'
        print(f'Start software...\n\t{url}\n')

        csrf_token = self.get_csrf_token()
        headers = {   
            #'Api-Key': self.api_key,
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cookie': authentication_cookie,
            'x-csrftoken': csrf_token,
        }

        ret = requests.post(url, headers=headers, json={"csrfmiddlewaretoken": csrf_token, "action": "post"})
        ret = ret.content.decode('utf8')
        print(ret)
        ret = json.loads(ret)
        assert ret['status'] == 0

    def get_csrf_token(self):
        ret = requests.get('https://www.tira.io/', headers={"Api-Key": self.api_key})

        return ret.content.decode('utf-8').split('name="csrfmiddlewaretoken" value="')[1].split('"')[0]

    def json_response(self, endpoint, params=None):
        cache_key = endpoint + '----' + ('' if not params else json.dumps(params))
        
        if cache_key in self.json_cache:
            return self.json_cache[cache_key]
        
        headers = {"Api-Key": self.api_key, "Accept": "application/json", "Api-Username": self.api_user_name}
        
        if self.api_key == 'no-api-key':
            del headers["Api-Key"]

        if self.api_user_name == 'no-api-key-user':
            del headers["Api-Username"]

        for i in range(self.failsave_retries):
            try:
                resp = requests.get(url='https://www.tira.io' + endpoint, headers=headers, params=params)
        
                if resp.status_code not in {200, 202}:
                    raise ValueError('Got statuscode ', resp.status_code, 'for ', endpoint, '. Got', resp)
                else:
                    break
            except Exception as e:
                sleep_time = 1+int(random()*self.failsave_max_delay)
                print(e)
                print(f'Code: {resp.status_code}')
                print(f'Error occured while fetching {endpoint}. I will sleep {sleep_time} seconds and continue.')
                time.sleep(sleep_time)

        self.json_cache[cache_key] = resp.json()

        return self.json_cache[cache_key]




from tira.ir_datasets_util import translate_irds_id_to_tirex
from tira.third_party_integrations import ensure_pyterrier_is_loaded
import pyterrier as pt


def test_for_tirex_id_robust04_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex( 'disks45-nocr-trec-robust-2004-20230209-training')
    
    assert expected == actual

def test_for_tirex_id_robust04_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex( 'disks45-nocr-trec-robust-2004-20230209-training')
    
    assert expected == actual

def test_for_tirex_id_cw09_2014_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'clueweb12-trec-web-2014-20230107-training'
    actual = translate_irds_id_to_tirex('clueweb12-trec-web-2014-20230107-training')
    
    assert expected == actual

def test_for_tirex_id_cw09_2014_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'clueweb12-trec-web-2014-20230107-training'
    actual = translate_irds_id_to_tirex('clueweb12-trec-web-2014-20230107-training')
    
    assert expected == actual

def test_for_raw_string_robust04_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex('disks45/nocr/trec-robust-2004')
    
    assert expected == actual

def test_for_raw_string_robust04_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex('disks45/nocr/trec-robust-2004')
    
    assert expected == actual

def test_for_raw_string_cw09_2014_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'clueweb12-trec-web-2014-20230107-training'
    actual = translate_irds_id_to_tirex('clueweb12/trec-web-2014')
    
    assert expected == actual

def test_for_raw_string_cw09_2014_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'clueweb12-trec-web-2014-20230107-training'
    actual = translate_irds_id_to_tirex('clueweb12/trec-web-2014')
    
    assert expected == actual

def test_for_pyterrier_dataset_robust04_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex(pt.get_dataset("irds:disks45/nocr/trec-robust-2004"))
    
    assert expected == actual

def test_for_pyterrier_dataset_robust04_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'disks45-nocr-trec-robust-2004-20230209-training'
    actual = translate_irds_id_to_tirex(pt.get_dataset("irds:disks45/nocr/trec-robust-2004"))
    
    assert expected == actual

def test_for_pyterrier_dataset_cw09_2009_01():
    ensure_pyterrier_is_loaded(patch_ir_datasets=False)
    expected = 'clueweb09-en-trec-web-2009-20230107-training'
    actual = translate_irds_id_to_tirex(pt.get_dataset("irds:clueweb09/en/trec-web-2009"))
    
    assert expected == actual

def test_for_pyterrier_dataset_cw09_2009_02():
    ensure_pyterrier_is_loaded(patch_ir_datasets=True)
    expected = 'clueweb09-en-trec-web-2009-20230107-training'
    actual = translate_irds_id_to_tirex(pt.get_dataset("irds:clueweb09/en/trec-web-2009"))
    
    assert expected == actual



from copy import deepcopy
from typing import NamedTuple
from tira.io_utils import all_lines_to_pandas
from tira.tirex import IRDS_TO_TIREX_DATASET, TIREX_DATASETS
import os

original_ir_datasets_load = None
try:
    import ir_datasets as original_ir_datasets
    original_ir_datasets_load = original_ir_datasets.load
except:
    pass


class TirexQuery(NamedTuple):
    query_id: str
    text: str
    title: str
    query: str
    description: str
    narrative: str

    def default_text(self):
        """
        title
        """
        return self.title

def register_dataset_from_re_rank_file(ir_dataset_id, df_re_rank, original_ir_datasets_id=None):
    """
    Load a dynamic ir_datasets integration from a given re_rank_file.
    The dataset will be registered for the id ir_dataset_id.
    The original_ir_datasets_id is used to infer the class of documents, qrels, and queries.
    """
    import ir_datasets
    from ir_datasets.datasets.base import Dataset
    original_dataset = ir_datasets.load(original_ir_datasets_id) if original_ir_datasets_id else None

    docs = __docs(df_re_rank, original_dataset)
    queries = __queries(df_re_rank, original_dataset)
    qrels = __qrels(df_re_rank, original_dataset)
    scoreddocs = __scored_docs(df_re_rank, original_dataset)

    dataset = Dataset(docs, queries, qrels, scoreddocs)
    ir_datasets.registry.register(ir_dataset_id, dataset)
    
    __check_registration_was_successful(ir_dataset_id, original_ir_datasets_id is None)


def translate_irds_id_to_tirex(dataset):
    if type(dataset) != str:
        if hasattr(dataset, 'irds_ref'):
            return translate_irds_id_to_tirex(dataset.irds_ref().dataset_id())
        else:
            raise ValueError(f'I can not handle {dataset}.')
    
    return IRDS_TO_TIREX_DATASET[dataset] if dataset in IRDS_TO_TIREX_DATASET else dataset


def __docs(df, original_dataset):
    from ir_datasets.formats import BaseDocs, GenericDoc

    class DynamicDocs(BaseDocs):
        def __init__(self, docs):
            self.docs = deepcopy(docs)

        def docs_iter(self):
            return deepcopy(docs).values().__iter__()

        def docs_count(self):
           return len(self.docs)

        def docs_store(self):
            return deepcopy(docs)

    docs = {}
    for _, i in df.iterrows():
        docs[i['docno']] = GenericDoc(doc_id=i['docno'], text= i['text'])

    return DynamicDocs(docs)


def __lazy_docs(tira_path):
    from ir_datasets.formats import BaseDocs, GenericDoc

    class DocumentsFromTira(BaseDocs):
        def __init__(self):
            self.docs = None
            if len(tira_path.split('/')) != 2:
                print(f'Please pass tira_path as <task>/<tira-dataset>. Got {tira_path}')
                return

            self.task, self.dataset = tira_path.split('/')

        def docs_iter(self):
            return self.get_docs().values().__iter__()

        def docs_count(self):
           return len(self.get_docs())

        def docs_store(self):
            return self.get_docs()

        def get_docs(self):
            if not self.docs:
                from tira.rest_api_client import Client as RestClient
                docs_file = RestClient().download_dataset(self.task, self.dataset)
                if os.path.isfile(docs_file + '/documents.jsonl.gz'):
                    docs_file = docs_file + '/documents.jsonl.gz'
                else:
                    docs_file = docs_file + '/documents.jsonl'
            
                self.docs = {}

                for _, i in all_lines_to_pandas(docs_file, False).iterrows():
                    self.docs[i['docno']] = GenericDoc(doc_id=i['docno'], text= i['text'])
                
            return self.docs

    ret = DocumentsFromTira()
    
    return ret


def __lazy_qrels(tira_path):
    from ir_datasets.formats import BaseQrels, TrecQrel, TrecQrels
    from ir_datasets.util.download import LocalDownload

    class QrelsFromTira(BaseQrels):
        def __init__(self):
            self.qrels = None
            if len(tira_path.split('/')) != 2:
                print(f'Please pass tira_path as <task>/<tira-dataset>. Got {tira_path}')
                return

            self.task, self.dataset = tira_path.split('/')

        def qrels_cls(self):
            return TrecQrel
        
        def qrels_iter(self):
            if not self.qrels:
                from tira.rest_api_client import Client as RestClient
                qrels_file = RestClient().download_dataset(self.task, self.dataset, truth_dataset=True) + '/qrels.txt'

                self.qrels = TrecQrels(LocalDownload(qrels_file), {})

            return self.qrels.qrels_iter()

        def qrels_defs(self):
            self.qrels_iter()
            return self.qrels.qrels_defs()

    return QrelsFromTira()


def __lazy_queries(tira_path):
    from ir_datasets.formats import BaseQueries

    class QueriesFromTira(BaseQueries):
        def __init__(self):
            self.queries = None
            if len(tira_path.split('/')) != 2:
                print(f'Please pass tira_path as <task>/<tira-dataset>. Got {tira_path}')
                return

            self.task, self.dataset = tira_path.split('/')

        def queries_cls(self):
            return TirexQuery
        
        def queries_iter(self):
            if not self.queries:
                from tira.rest_api_client import Client as RestClient
                queries_file = RestClient().download_dataset(self.task, self.dataset, truth_dataset=True) + '/queries.jsonl'

                self.queries = {}
                for _, i in all_lines_to_pandas(queries_file, False).iterrows():
                    orig_query = None if 'original_query' not in i else i['original_query']
                    
                    description = None if (not orig_query or type(orig_query) != dict or 'description' not in orig_query) else orig_query['description']
                    narrative = None if (not orig_query or type(orig_query) != dict or 'narrative' not in orig_query) else orig_query['narrative']
                    self.queries[i['qid']] = TirexQuery(query_id=i['qid'], text=i['query'], query=i['query'], title=i['query'], description=description, narrative=narrative)

            return deepcopy(self.queries).values().__iter__()
    
    return QueriesFromTira()

def __queries(df, original_dataset):
    from ir_datasets.formats import BaseQueries

    class DynamicQueries(BaseQueries):
        def __init__(self, queries):
            self.queries = deepcopy(queries)

        def queries_iter(self):
            return deepcopy(self.queries).values().__iter__()

        def queries_cls(self):
            return TirexQuery

    queries = {}
    for _, i in df.iterrows():
        i = i.to_dict()
        orig_query = None if 'original_query' not in i else i['original_query']
        description = None if (not orig_query or type(orig_query) != dict or 'description' not in orig_query) else orig_query['description']
        narrative = None if (not orig_query or type(orig_query) != dict or 'narrative' not in orig_query) else orig_query['narrative']
        queries[i['qid']] = TirexQuery(query_id=i['qid'],  text=i['query'], query=i['query'], title=i['query'], description=description, narrative=narrative)

    return DynamicQueries(queries)



def __qrels(path_to_re_rank_file, original_dataset):
    from ir_datasets.formats import BaseQrels

    if not original_dataset:
        return None

    class DynamicQrels(BaseQrels):
        def __init__(self, qrels):
            self.qrels = list(original_dataset.qrels.qrels_iter())
        
        def qrels_iter(self):
            return self.qrels.__iter__()

    return DynamicQrels


def __scored_docs(df, original_dataset):
    from ir_datasets.formats import BaseScoredDocs
    
    class GenericScoredDocWithRank(NamedTuple):
        query_id: str
        doc_id: str
        score: float
        rank: int

    class DynamicScoredDocs(BaseScoredDocs):
        def __init__(self, docs):
            self.docs = docs

        def scoreddocs_iter(self):
            return deepcopy(self.docs).__iter__()

    docs = []

    for _, i in df.iterrows():
        docs += [GenericScoredDocWithRank(i['qid'], i['docno'], i['score'], i['rank'])]
    
    return DynamicScoredDocs(docs)


def static_ir_dataset(directory, existing_ir_dataset=None):
    from ir_datasets.datasets.base import Dataset
    import pandas as pd
    if existing_ir_dataset is None:
        queries_file = directory + '/queries.jsonl'
        docs_file = directory
        if os.path.isfile(docs_file + '/documents.jsonl.gz'):
            docs_file = docs_file + '/documents.jsonl.gz'
        else:
            docs_file = docs_file + '/documents.jsonl'

        docs = __docs(all_lines_to_pandas(docs_file, False) if os.path.isfile(docs_file) else pd.DataFrame([]), None)
        queries = __queries(all_lines_to_pandas(queries_file, False), None)
        return static_ir_dataset(directory, Dataset(docs, queries))

    class StaticIrDatasets():
        def load(self, dataset_id):
            print(f'Load ir_dataset from "{directory}" instead of "{dataset_id}" because code is executed in TIRA.')
            return existing_ir_dataset
        def topics_file(self, dataset_id):
            return directory + '/queries.xml'
    
    return StaticIrDatasets()


def ir_dataset_from_tira_fallback_to_original_ir_datasets():
    from ir_datasets.datasets.base import Dataset

    class IrDatasetsFromTira():
        def load(self, dataset_id):
            try:
                return original_ir_datasets_load(dataset_id)
            except:
                print(f'Load ir_dataset "{dataset_id}" from tira.')
                docs = self.lazy_docs(dataset_id)
                queries = self.lazy_queries(dataset_id)
                qrels = self.lazy_qrels(dataset_id)
                return Dataset(docs, queries, qrels)

        def topics_file(self, tira_path):
            from tira.rest_api_client import Client as RestClient
            task, dataset = tira_path.split('/')
            return RestClient().download_dataset(task, dataset, truth_dataset=True) + '/queries.xml'

    ret = IrDatasetsFromTira()
    ret.lazy_docs = __lazy_docs
    ret.lazy_queries = __lazy_queries
    ret.lazy_qrels = __lazy_qrels

    return ret

def __check_registration_was_successful(ir_dataset_id, ignore_qrels=True):
    import ir_datasets
    dataset = ir_datasets.load(ir_dataset_id)

    assert dataset.has_docs(), "dataset has no documents"
    assert dataset.has_queries(), "dataset has no queries"
    assert ignore_qrels or dataset.has_qrels(), "dataset has no qrels"
    assert dataset.has_scoreddocs(), "dataset has no scored_docs"

